# Node-level operations/abstractions
import logging

import binascii
import bz2
from collections import OrderedDict
import importlib
import sys

from .flash import _get_file_path_for_version

from distutils.sysconfig import get_python_lib

# Helpers
import formatters
from util import first_or_none

import os

from tornado.concurrent import Future
from tornado.gen import coroutine, sleep

from snapconnect import snap

from snaplib import ScriptsManager

from snappy import SnapConstantsHelper

from .util import fix_up_target, CRYPTO_LOOKUP
from .gus import NoResponseError

log = logging.getLogger(__name__)


@coroutine  # noqa
@fix_up_target
def get_info(gus, data, target, verbosity=0):
    """Get node info from the target node.

    C.f. Portal's Node Info and Parameters pane.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node
        verbosity (int): How deep do we go?

    Returns:
        None, though data will be populated.

    """

    data['_errors'] = 0

    @coroutine
    def query_nvparams(nvparams):
        """
        This helper function queries multiple nvparams. The

        Args:
            nvparams (dict): Dictionary of nvparam definitions; the key is the
                nvparam id and the value is the key under which the value will
                be stored in data

        Returns:
            None, though data will be populated.
        """

        for nvparam_id, data_key in nvparams.iteritems():
            try:
                val = yield gus.callback_ucast_rpc(target, "loadNvParam", nvparam_id)
            except NoResponseError:
                val = None

            if val is None:
                data['_errors'] += 1
            data[data_key] = first_or_none(val)

    try:
        val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_MAC_ADDR_ID)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['address'] = binascii.hexlify(val)
    else:
        data['address'] = None

    try:
        val = yield gus.callback_ucast_rpc(target, "getChannel")
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['current_channel'] = first_or_none(val)
    else:
        data['current_channel'] = None

    try:
        val = yield gus.callback_ucast_rpc(target, "getNetId")
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['current_netid'] = "0x%04x" % (val & 0xffff)
    else:
        data['current_netid'] = None

    try:
        val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_CHANNEL_ID)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    data['nv_channel'] = first_or_none(val)

    try:
        val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_NETWORK_ID)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['nv_net_id'] = "0x%04x" % (val & 0xffff)
    else:
        data['nv_net_id'] = None

    try:
        val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_FEATURE_BITS_ID)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['nv_feature_bits'] = "0x%04x" % (val & 0xffff)
    else:
        data['nv_feature_bits'] = None

    try:
        val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_VENDOR_SETTINGS_ID)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1
    val = first_or_none(val)
    if val is not None:
        data['nv_vendor_settings'] = "0x%04x" % (val & 0xffff)
    else:
        data['nv_vendor_settings'] = None

    if verbosity > 0:
        try:
            val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_DEVICE_NAME_ID)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        data['nv_device_name'] = first_or_none(val)

        try:
            val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_DEVICE_TYPE_ID)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        data['nv_device_type'] = first_or_none(val)

        # Get more info
        try:
            val = yield gus.callback_ucast_rpc(target, "imageName")
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        data['script_name'] = first_or_none(val)

        try:
            val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_SNAPPY_CRC_ID)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        val = first_or_none(val)
        if val is not None:
            data['script_crc'] = "0x%04x" % (val & 0xffff)
        else:
            data['script_crc'] = None

        try:
            val = yield gus.callback_ucast_rpc(target, "vmStat", SnapConstantsHelper.VM_SPACE, callback_name="tellVmStat")
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        if val is not None:
            (code, val) = val
            if code & 0xFF == SnapConstantsHelper.VM_SPACE:
                data['space_available'] = val & 0xffff
            else:
                data['space_available'] = None
        else:
            data['space_available'] = None

        try:
            val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_SYS_PLATFORM_ID)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
        data['platform'] = first_or_none(val)

        version = {}
        for i in range(5, 9):
            try:
                val = yield gus.callback_ucast_rpc(target, "getInfo", i)
            except NoResponseError:
                val = None

            if val is None:
                data['_errors'] += 1
            version[i] = first_or_none(val)

        version[8] = CRYPTO_LOOKUP[version[8]] if version[8] is not None else "Unknown"

        data['firmware_version'] = "{}.{}.{} with {} encryption".format(
            version[5], version[6], version[7], version[8])

    if verbosity > 1:
        # Get even more info: Radio, UART, Multicast
        radio_nvparams = {
            SnapConstantsHelper.NV_SNAP_MAX_RETRIES_ID: 'radio_unicast_retries',
            SnapConstantsHelper.NV_SNAP_LQ_THRESHOLD_ID: 'radio_lq_threshold',
        }
        yield query_nvparams(radio_nvparams)

        uart_nvparams = {
            SnapConstantsHelper.NV_DEFAULT_UART_ID: 'uart_default_uart',
            SnapConstantsHelper.NV_UART_DM_TIMEOUT_ID: 'uart_dm_timeout',
            SnapConstantsHelper.NV_UART_DM_THRESHOLD_ID: 'uart_dm_threshold',
            SnapConstantsHelper.NV_UART_DM_INTERCHAR_ID: 'uart_dn_interchar',
            SnapConstantsHelper.NV_UART0_BAUDRATE_ID: 'uart_uart0_baudrate',
            SnapConstantsHelper.NV_UART1_BAUDRATE_ID: 'uart_uart1_baudrate',
        }
        yield query_nvparams(uart_nvparams)

        mcast_nvparams = {
            SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID: 'mcast_group_interest_mask',
            SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID: 'mcast_group_forwarding_mask',
            SnapConstantsHelper.NV_GROUP_SERIAL_FORWARDING_MASK_ID: 'mcast_serial_forwarding_mask',
            SnapConstantsHelper.NV_CARRIER_SENSE_ID: 'mcast_carrier_sense',
            SnapConstantsHelper.NV_COLLISION_DETECT_ID: 'mcast_collision_detect',
            SnapConstantsHelper.NV_COLLISION_AVOIDANCE_ID: 'mcast_collision_avoidance',
            SnapConstantsHelper.NV_CS_CD_LEVEL_ID: 'mcast_cs_cd_level',
            SnapConstantsHelper.NV_CSMA_TIMESLOT_SETTINGS_ID: 'mcast_csma_timeslot_settings',
        }
        yield query_nvparams(mcast_nvparams)

    if verbosity > 2:
        # Get all the info: Mesh
        mesh_nvparams = {
            SnapConstantsHelper.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID: 'mesh_route_max_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID: 'mesh_route_min_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_NEW_TIMEOUT_ID: 'mesh_route_new_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_USED_TIMEOUT_ID: 'mesh_route_used_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_DELETE_TIMEOUT_ID: 'mesh_route_delete_timeout',
            SnapConstantsHelper.NV_MESH_RREQ_TRIES_ID: 'mesh_rreq_tries',
            SnapConstantsHelper.NV_MESH_RREQ_WAIT_TIME_ID: 'mesh_wait_time',
            SnapConstantsHelper.NV_MESH_INITIAL_HOPLIMIT_ID: 'mesh_initial_hop_limit',
            SnapConstantsHelper.NV_MESH_MAX_HOPLIMIT_ID: 'mesh_max_hop_limit',
            SnapConstantsHelper.NV_MESH_SEQUENCE_NUMBER_ID: 'mesh_seq_number',
            SnapConstantsHelper.NV_MESH_OVERRIDE_ID: 'mesh_override',
            SnapConstantsHelper.NV_MESH_PENALTY_LQ_THRESHOLD_ID: 'mesh_penalty_lq_threshold',
            SnapConstantsHelper.NV_MESH_REJECTION_LQ_THRESHOLD_ID: 'mesh_rejection_lq_threshold',
        }
        yield query_nvparams(mesh_nvparams)


@coroutine
@fix_up_target
def reboot(gus, data, target, delay=None):
    """Reboot the target node.

    Equivalent of the reboot(delay) command.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node
        delay (Optional[int]): (Optional) ms to wait before rebooting

    Returns:
        None

    """
    if delay is None:
        val = yield gus.callback_ucast_rpc(target, "reboot")
    else:
        val = yield gus.callback_ucast_rpc(target, "reboot", delay)
    data['reboot'] = first_or_none(val)


@coroutine
@fix_up_target
def script_info(gus, data, target):
    """Get the script name and CRC from the target node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node

    Returns:
        None

    """
    val = yield gus.callback_ucast_rpc(target, "imageName")
    data['script_name'] = first_or_none(val)

    val = yield gus.callback_ucast_rpc(target, "loadNvParam", snap.NV_SNAPPY_CRC_ID)
    data['script_crc'] = first_or_none(val)


@coroutine
@fix_up_target
def script_erase(gus, data, target):
    """OTA Script erase from the target node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node

    Returns:
        None

    """
    yield gus.callback_ucast_rpc(target, 'eraseImage')


stat_lookup = OrderedDict([
    ('info.vendor', {'id': 0, 'rpc': 'getInfo', 'formatter': formatters.format_as_vendor}),
    ('info.primary-communications-interface', {'id': 1, 'rpc': 'getInfo', 'formatter': formatters.format_as_interface}),
    ('info.cpu', {'id': 2, 'rpc': 'getInfo', 'formatter': formatters.format_as_cpu}),
    ('info.platform', {'id': 3, 'rpc': 'getInfo', 'formatter': formatters.format_as_platform}),
    ('info.build', {'id': 4, 'rpc': 'getInfo', 'formatter': formatters.format_as_build}),

    ('info.version.major', {'id': 5, 'rpc': 'getInfo'}),
    ('info.version.minor', {'id': 6, 'rpc': 'getInfo'}),
    ('info.version.build', {'id': 7, 'rpc': 'getInfo'}),

    ('info.rpc.buffer', {'id': 9, 'rpc': 'getInfo'}),
    ('info.rpc.is-multicast', {'id': 10, 'rpc': 'getInfo'}),
    ('info.rpc.remaining-ttl', {'id': 11, 'rpc': 'getInfo'}),
    ('info.rpc.is-dmcast', {'id': 24, 'rpc': 'getInfo'}),

    ('info.mcast.groups', {'id': 27, 'rpc': 'getInfo'}),

    ('info.dmcast.address-index', {'id': 26, 'rpc': 'getInfo'}),
    ('info.dmcast.original-ttl', {'id': 28, 'rpc': 'getInfo'}),

    ('info.strings.tiny-left', {'id': 19, 'rpc': 'getInfo'}),
    ('info.strings.small-left', {'id': 12, 'rpc': 'getInfo'}),
    ('info.strings.medium-left', {'id': 13, 'rpc': 'getInfo'}),
    ('info.strings.large-left', {'id': 20, 'rpc': 'getInfo'}),

    ('info.script.is-first-execution', {'id': 21, 'rpc': 'getInfo'}),
    ('info.script.base-address', {'id': 22, 'rpc': 'getInfo'}),
    ('info.script.base-bank', {'id': 23, 'rpc': 'getInfo'}),

    ('info.encryption-available', {'id': 8, 'rpc': 'getInfo', 'formatter': formatters.format_as_encryption_status}),
    ('info.route-table-size', {'id': 14, 'rpc': 'getInfo'}),
    ('info.route-table-entries', {'id': 15, 'rpc': 'getInfo'}),
    ('info.bank-free-space', {'id': 16, 'rpc': 'getInfo'}),
    ('info.stdin-hook-trigger', {'id': 18, 'rpc': 'getInfo'}),
    ('info.read-and-reset-delay-factor', {'id': 25, 'rpc': 'getInfo'}),

    ('stat.ds-null.tx-buffers', {'id': 0, 'rpc': 'getStat'}),
    ('stat.ds-uart0.rx-buffers', {'id': 1, 'rpc': 'getStat'}),
    ('stat.ds-uart0.tx-buffers', {'id': 2, 'rpc': 'getStat'}),
    ('stat.ds-uart1.rx-buffers', {'id': 3, 'rpc': 'getStat'}),
    ('stat.ds-uart1.tx-buffers', {'id': 4, 'rpc': 'getStat'}),
    ('stat.ds-transparent.rx-buffers', {'id': 5, 'rpc': 'getStat'}),
    ('stat.ds-transparent.tx-buffers', {'id': 6, 'rpc': 'getStat'}),
    ('stat.packet-serial.rx-buffers', {'id': 7, 'rpc': 'getStat'}),
    ('stat.packet-serial.tx-buffers', {'id': 8, 'rpc': 'getStat'}),
    ('stat.radio.rx-buffers', {'id': 9, 'rpc': 'getStat'}),
    ('stat.radio.tx-buffers', {'id': 10, 'rpc': 'getStat'}),
    ('stat.radio.forwarded-unicasts', {'id': 11, 'rpc': 'getStat'}),
    ('stat.packet-serial.forwarded-unicasts', {'id': 12, 'rpc': 'getStat'}),
    ('stat.radio.forwarded-xcasts', {'id': 13, 'rpc': 'getStat'}),
    ('stat.packet-serial.forwarded-xcasts', {'id': 14, 'rpc': 'getStat'}),
])


@coroutine
@fix_up_target
def get_stats(gus, data, target, stat):
    """Get stats and info about the node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node
        stat (str): The name of the stat to query

    Returns:
        None

    """
    try:
        val = None

        if stat in stat_lookup:
            stat_info = stat_lookup[stat]
            val = yield gus.callback_ucast_rpc(target, stat_info['rpc'], stat_info['id'])
            data[stat] = stat_info.get('formatter', formatters.format_noop)(first_or_none(val))
        elif stat == 'info':
            # Get all info stats
            for stat in stat_lookup:
                if stat.startswith('info'):
                    yield get_stats(gus, data, target, stat)
        elif stat == 'stat':
            # Get all stat stats
            for stat in stat_lookup:
                if stat.startswith('stat'):
                    yield get_stats(gus, data, target, stat)
        elif stat == 'all':
            # Get all info and stat stats
            for stat in stat_lookup:
                yield get_stats(gus, data, target, stat)
        else:
            raise KeyError()
    except KeyError:
        raise ValueError('%s is not a valid stat name' % stat)


@coroutine
@fix_up_target
def firmware_info(gus, data, target):
    """Get the firmware version info from the target node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        data (OrderedDict()): Dictionary for result data
        target (str, SNAP Address): Target Node

    Returns:
        None

    """
    version = {}
    for i in range(5, 9):
        val = yield gus.callback_ucast_rpc(target, "getInfo", i)
        version[i] = first_or_none(val)

    version[8] = CRYPTO_LOOKUP[version[8]] if version[8] is not None else "Unknown"

    data['firmware_version'] = "{}.{}.{} with {} encryption".format(
        version[5], version[6], version[7], version[8])


@coroutine
@fix_up_target
def get_vmstat(gus, data, target, vmstat_code, vmstat_arg):
    """Get the specified VMStat from the target.

    Equivalent of vmStat(code, arg)

    Args:

    Returns:
        None

    """
    val = yield gus.callback_ucast_rpc(target,
                                       'vmStat',
                                       vmstat_code, vmstat_arg,
                                       callback_name='tellVmStat')

    (word, d) = val

    word = word & 0xFFFF

    code = word & 0xFF
    hibyte = word >> 8

    data['code'] = code
    data['hibyte'] = hibyte
    data['data'] = d


@coroutine
@fix_up_target
def traceroute(gus, data, target):
    future = Future()

    def on_traceroute(node_addr, round_trip_time, hops):
        data['node'] = node_addr
        data['round_trip_time'] = round_trip_time
        data['hops'] = hops

        # Mark the traceroute operation as completed so that the coroutine can exit
        future.set_result(None)

    gus.on_traceroute(on_traceroute)

    gus.traceroute(target)

    # Wait for the future to resolve before exiting the coroutine
    yield future


@coroutine
@fix_up_target
def scan_energy(gus, data, target, num=1, delay=0):

    for i in xrange(num):
        val = yield gus.callback_ucast_rpc(target, 'scanEnergy')
        energies = first_or_none(val)

        # Unpack the response
        for i in xrange(len(energies)):
            data.setdefault('energy_channel_%02d' % i, []).append(ord(energies[i]))

        log.debug("DATA: %r", data)
        yield sleep(delay)
    else:
        for channel, scans in data.iteritems():
            data[channel] = float(sum(scans)) / len(scans)

    log.debug("DATA: %r", data)


@coroutine
@fix_up_target
def start_output_interceptor(gus, data, target, intercept_stdout=True, intercept_stderr=True):
    if intercept_stdout:
        yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_STDIO)
    if intercept_stderr:
        yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_ERROR)
    yield gus.callback_ucast_rpc(target, 'ucastSerial', binascii.unhexlify(gus.address))


@coroutine
@fix_up_target
def stop_output_interceptor(gus, data, target):
    yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_STDIO)
    yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_ERROR)


def lookup_platform_by_id(module_id, platform_str=None):
    log.debug("Looking up module id %d, str %r", module_id, platform_str)

    candidates = []
    for k, v in SnapConstantsHelper.platforms_strings_to_ids.iteritems():
        if v == module_id:
            log.debug("Found a candidate: %s", k)
            candidates.append(k)
            # If it's an exact match, return that.
            if platform_str in SnapConstantsHelper.platforms_strings_to_ids and SnapConstantsHelper.platforms_strings_to_ids[platform_str] == module_id:
                return platform_str

    log.debug("Candidates: %r", candidates)

    if candidates:
        return candidates[0]
    else:
        return None


def lookup_platform_by_core_and_id(core, module_id, platform_str=None):
    log.debug("Looking up module part (to use with SNAPbuild) for core %s, module id %d, str %r", core, module_id, platform_str)

    # Extend the path so we can find firmware support packages
    sys.path.append(get_python_lib())

    # This is a check for "Are we being executed under an active virtualenv?"
    # Different from "Are we installed in a virtualenv"
    if 'VIRTUAL_ENV' in os.environ:
        sys.path.append(os.path.abspath(get_python_lib(prefix=os.environ['VIRTUAL_ENV'])))

    try:
        # Import the firmware metadata based on the version
        firmware_pkg = "snap_firmware_" + core.replace('.', '_')
        core_module = importlib.import_module(firmware_pkg + ".metadata")
    except:
        core_module = None

    if core_module is None:
        log.error("Couldn't find firmware support package for core '%s', you may need to install it " +
                  "by running the following command:\n" +
                  "    pip install -i https://update.synapse-wireless.com/pypi snap-firmware-%s",
                  core, core)
        return None

    # We rely on the snap-firmware-* metadata to figure out what a valid "module part" is now.
    # This sort of stuff *really* ought to be burned into the modules at MFG time,
    # but it isn't as of 2016-03-28.
    # So! What we're going to do here is a rough inverse of the lookup SNAPbuild will do when
    # it needs the part to get back to the CPU.
    # For C code support...we'll just have to hope that the antenna variants never have differing
    # linker offsets...

    # What we need is a "module_id" to "part number" mapping for the particular core.

    candidates = []
    supported_modules = core_module.get_supported_modules()

    all_metadata = {module: core_module.get_metadata(module, core, True) for module in supported_modules}

    log.debug("All Metadata: %r", all_metadata)

    for module, metadata in all_metadata.iteritems():
        if metadata['platform'] == module_id:
            candidates.append((module, core_module.get_module_platform_string(module), metadata))

    candidates = sorted(candidates)

    log.debug("Candidates: %r", candidates)

    if candidates:
        log.warning("Multiple parts (%s) match module ID %d", ",".join([c[0] for c in candidates]), module_id)
        log.info("Attempting to match by platform NV Param %s...", platform_str)
        for c in candidates:
            if c[1] == platform_str:
                log.info("Found a match by platform NV Param: %r", c)
                return c[0]

        log.warning("No match by platform NV Param %s found, falling back on first candidate: %r", candidates[0])
        return candidates[0][0]
    else:
        return None


@coroutine
@fix_up_target
def get_node_core_module(gus, data, target):
    """Attempts to back out the Module Part info from vmStat(6).

    Figuring out the module part from the node itself is a bit of guesswork.
    The values returned here are sufficient for SNAPbuild, but may not be an exact match for your modules.
    If you *really* need an exact match, you should be using something like the platform NV param in your scripts.
    """

    attempts = 0

    while attempts < 3:
        attempts += 1

        val = yield gus.callback_ucast_rpc(target, 'vmStat', 6, callback_name='tellVmStat')

        (word, d) = val

        word = word & 0xFFFF
        code = word & 0xFF

        if code == 6:
            info = d.split(',')
            if len(info) >= 8:
                core = info[0]
                module_id = int(info[5])
                platform_str = info[7]
                module = lookup_platform_by_core_and_id(core, module_id, platform_str=platform_str)

                data['core'] = core
                data['module'] = module
                data['platform'] = platform_str
                break


@coroutine
@fix_up_target
def ping(gus, data, target):
    result = {}
    yield get_vmstat(gus, result, target, 5, None)
    data['lq'] = result['hibyte']


@coroutine
@fix_up_target
def ota_script_upload(gus, data, target, filename, progress_cb=None, finished_cb=None):
    """Futures-compatible OTA Script upload"""

    # Read the image from the filename
    with open(filename, "rb") as f:
        image = ScriptsManager.getSnappyStringFromExport(f.read())

    if not image:
        raise Exception("Unable to load SNAPpy image from %s", filename)

    # Determine chunk size, allow for RPC_CRC and Packet CRC
    chunk_size = SnapConstantsHelper.getMaxPacketSize() - 30

    # Chunk the image
    def chunk(array, size):
        """Generator that yields sized chunks from array."""
        for i in xrange(0, len(array), size):
            yield array[i:i + size]

    # Send eraseImage
    yield gus.callback_ucast_rpc(target, 'eraseImage')

    image_size = len(image)

    # Send the chunks
    offset = 0
    for i, c in enumerate(chunk(image, chunk_size)):
        result = yield gus.callback_ucast_rpc(target, 'writeChunk', offset, c, callback_name='tellVmStat')
        if result is None:
            raise Exception("Unable to load chunk %d on target %s", i, target)

        (word, d) = result

        word = word & 0xFFFF
        code = word & 0xFF

        if code != 2:
            log.warn("Unexpected tellVmStat from %s", target)

        written = d & 0xFFFF

        if offset != written:
            raise Exception("Failed to write chunk at %d, written %d", offset, written)

        offset += len(c)

        if progress_cb:
            progress_cb(offset, image_size)

    # Send a reboot
    result = yield gus.callback_ucast_rpc(target, 'reboot')
    if result is None:
        raise Exception("Failed to reboot target %s", target)

    data['spy'] = filename

    if finished_cb:
        finished_cb()


def _get_file_architecture(filename):
    """Return the (file_cpu, file_platform) for the given filename

    If the type cannot be determined then it will be set to None
    """
    file_cpu, file_platform = None, None
    if "MC1322x" in filename or "SM700" in filename:
        file_cpu = SnapConstantsHelper.CPU_ARM7
    elif "STM32W108" in filename:
        file_cpu = SnapConstantsHelper.CPU_ARM_CORTEX_M3
    # ATmega family
    elif "SM222" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM222_PFM
    elif "ATmega128RFR2" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega128RFR2_PFM
    elif "ATmega1284RFR2" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega1284RFR2_PFM
    elif "SM220" in filename or "RF220UF1" in filename or "SN220" in filename:
        # RF220UF1 is an SM220UF1 on an RF form factor carrier board
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM220UF1_PFM
    elif "RF220SU" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF220_SU
    elif "SM200" in filename or "RF200PU1" in filename or "RF200P81" in filename:
        # RF200PU1 is an SM200 module on an RF module form factor:
        # So is RF200P81
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM200_PFM
    elif "SS200" in filename or "RF200" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF200_PFM
    elif "RF266" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF266_PFM
    elif "ATmega128RFA1" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega128RFA1

    return file_cpu, file_platform


def _get_layout(cpu):
    """We only really support the ATmega128RFA1 for OTA firmware upgrades"""
    from snaplayout import ATmega128RFA1_Layout
    layout_map = {SnapConstantsHelper.CPU_ATMEGA: ATmega128RFA1_Layout.ATmega128RFA1_Layout}

    return layout_map.get(cpu, None)


def _get_layout_from_filename(filename):
    file_cpu, file_platform = _get_file_architecture(filename)
    layout_class = _get_layout(file_cpu)

    log.info("Firmware file %s is for CPU %s, Platform %s, will use layout %s", filename, file_cpu, file_platform, layout_class)

    # Read the image from the filename
    with open(filename, "rb") as f:
        contents = f.read()

        # *.sfi files are compressed
        if os.path.splitext(filename)[1] == '.sfi':
            contents = bz2.decompress(contents)

    layout = layout_class(contents)

    if not layout:
        raise Exception("Unable to load Firmware image from %s" % filename)

    return layout


@coroutine
@fix_up_target
def ota_firmware_upload(gus, data, target, core, filename=None, progress_cb=None, finished_cb=None):
    """Futures-compatible OTA Firmware upload"""

    # Determine target's core/module type
    yield get_node_core_module(gus, data, target)

    module = data['module']

    log.info("Target %s is a module %s, currently running core %s", target, module, data['core'])

    if filename is None:

        # Extend the path so we can find firmware support packages
        sys.path.append(get_python_lib())

        # This is a check for "Are we being executed under an active virtualenv?"
        # Different from "Are we installed in a virtualenv"
        if 'VIRTUAL_ENV' in os.environ:
            sys.path.append(os.path.abspath(get_python_lib(prefix=os.environ['VIRTUAL_ENV'])))

        try:
            metadata_module_name = "snap_firmware_%s" % (core.replace('.', '_'),)
            metadata = importlib.import_module(metadata_module_name + '.metadata')
            log.debug("Looking for firmware for %s", metadata_module_name)

            filename = _get_file_path_for_version(metadata, module, core)
        except:
            log.exception("Couldn't find firmware module %s, pip install %s", metadata_module_name)
            raise
    else:
        # A specific filename was provided, try it.
        filename = os.path.abspath(os.path.join(os.path.curdir, os.path.expanduser(filename)))

    if data['module'] not in filename:
        raise Exception("%s isn't valid for target %s, expecting a firmware for %s" % (filename, target, data['module']))

    log.info("Loading firmware %s onto module %s", filename, target)

    layout = _get_layout_from_filename(filename)

    # Determine chunk size, allow for RPC_CRC and Packet CRC
    chunk_size = SnapConstantsHelper.getMaxPacketSize() - 30

    # Chunk the image
    def chunk(array, size):
        """Generator that yields sized chunks from array."""
        for i in xrange(0, len(array), size):
            yield array[i:i + size]

    # Get Bank Free Space
    result = yield gus.callback_ucast_rpc(target, 'getInfo', 16)
    if result is None:
        raise Exception("Couldn't get bank free space on %s" % target)

    free_space = result[0] & 0xFFFF

    if free_space == 0:
        raise Exception("Node %s reports it doesn't support OTA Firmware Upgrades." % target)

    image = layout.getContents(free_space)

    # Send eraseImage
    yield gus.callback_ucast_rpc(target, 'eraseImage')

    image_size = len(image)

    # Send the chunks
    offset = 0
    for i, c in enumerate(chunk(image, chunk_size)):
        result = yield gus.callback_ucast_rpc(target, 'writeChunk', offset, c, callback_name='tellVmStat')
        if result is None:
            raise Exception("Unable to load chunk %d on target %s" % (i, target))

        (word, d) = result

        word = word & 0xFFFF
        code = word & 0xFF

        written = d & 0xFFFF

        if code != 2:
            log.warn("Unexpected tellVmStat from %s", target)

        if written != offset:
            raise Exception("Failed to write chunk at %d, written %d", offset, written)

        offset += len(c)

        if progress_cb:
            progress_cb(os.path.basename(filename), offset, image_size)

    # Send a reboot
    result = yield gus.callback_ucast_rpc(target, 'reboot')
    if result is None:
        raise Exception("Failed to reboot target %s" % target)

    data['firmware'] = filename

    if finished_cb:
        finished_cb()
