import binascii
import logging
import pprint
import Queue
import timeit

from snaplib import RpcCodec
from snappy import SnapConstantsHelper
from snaplib import TraceRouteCodec
from snaplib import PacketSerialProtocol
from snaplib import Snap2Decoder
from snaplib import pme
try:
    from snaplib import aes
    AES_SUPPORT = True
except ImportError:
    AES_SUPPORT = False
from serialwrapper import PyserialDriver

import tornado.ioloop

from sniffer_event import SnifferEvent

START_CMD = 'E'
STOP_CMD = 'D'

log = logging.getLogger(__name__)


class SnifferException(Exception):
    pass


class Sniffer(object):
    def __init__(self, pathToLibrary='.'):
        self.timer_callbacks = []
        self.event_callbacks = []
        self.pyserial_driver = \
            PyserialDriver.PyserialWrapper(openCallback=self.on_conn_open,
                                           dllPath=pathToLibrary)
        self.pyserial_driver.registerCloseCallback(self.on_comm_close)
        self.pyserial_driver.BAUDRATE = 115200
        self.pkt_serial = \
            PacketSerialProtocol.PacketSerialDriver(self.pyserial_driver, self)
        self.pkt_serial.decode = False
        if __debug__:
            self.pkt_serial.verboseLogging = True
        self.pkt_serial.registerCallback("rx", self.on_data)
        self.pkt_serial.neverAck = True
        self.pkt_serial.check_len = False
        self.data_queue = Queue.Queue()
        self.last_ms = None
        self.last_timestamp = None
        self.last_seq = None

        self.periodic_callback = tornado.ioloop.PeriodicCallback(self.on_timer, 100)
        self.periodic_callback.start()

        tornado.ioloop.IOLoop.current().add_callback(self.poll)

        if AES_SUPPORT:
            self.crypto = Snap2Decoder.crypto = aes.CipherStealingAES128("")
        else:
            self.crypto = None
        self._radiorate = 0
        self._lastfind = ""

        self.crypto_key = ""
        self.crypto_type = SnapConstantsHelper.ENCRYPTION_TYPE_NONE
        self.crypto_enabled = False

    def on_event(self, callback):
        if callback not in self.event_callbacks:
            self.event_callbacks.append(callback)

    def registerPollCallback(self, callback):
        """PacketSerialDriver wants this."""
        log.debug("called registerPollCallback()")
        assert callable(callback)
        if callable(callback):
            self.timer_callbacks.append(callback)

    def poll(self):
        if self:  # Check for PyDeadObjectError
            try:
                self.pyserial_driver.readPoll()
                for cb in self.timer_callbacks:
                    cb()
                self.pyserial_driver.writePoll()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("A poll error occurred")
                if __debug__:
                    raise

        tornado.ioloop.IOLoop.current().add_callback(self.poll)

    def on_timer(self):
        log.debug("on_timer")
        processed = 0
        try:
            while self:
                log.debug("checking for events in the queue")
                (timestamp, data, lq) = self.data_queue.get_nowait()
                log.debug("got an event")
                self.process_data(timestamp, data, lq)
                log.debug("processed event")
                processed += 1
        except Queue.Empty:
            pass

        if processed:
            pass

    def on_conn_open(self, portstr='', baud=0):
        log.debug("on_conn_open(portstr='%s', baud=%d", portstr, baud)

    def on_comm_close(self, *args):
        log.debug("on_comm_close(args=%s)", args)

    def check_for_dropped_packets(self, seq):
        """Return the number of dropped packets, or zero if none dropped."""
        if self.last_seq is not None:
            next_seq = (self.last_seq + 1) % 2 ** 16
            if seq > next_seq:
                return seq - next_seq
        return 0

    def adjust_timestamp(self, curr_timestamp, ms_timestamp):
        if self.last_ms is None or curr_timestamp - self.last_timestamp > 0xFFFF * 0.001:
            timestamp = curr_timestamp
        else:
            if ms_timestamp < self.last_ms:
                ms_timestamp += 0xffff
            timestamp = self.last_timestamp + (ms_timestamp - self.last_ms) * 0.001
        return timestamp

    def enqueue_data(self, timestamp, data):
        if len(data) > 4:
            log.debug("Adding data to the queue")
            self.data_queue.put_nowait((timestamp, data[:-4], ord(data[-2:-1])))
        elif __debug__:
            log.debug("Received too small of a packet")

    def on_data(self, data, intf=None):
        log.debug("on_data(data=%r, intf=%s", data, intf)

        if len(data) <= 6:
            log.debug("Received too small of a packet")
            return

        seq = ord(data[-1])
        self.check_for_dropped_packets(seq)

        curr_timestamp = timeit.default_timer()
        ms_timestamp = int(binascii.hexlify(data[-4:-2]), 16) % 2 ** 16
        timestamp = self.adjust_timestamp(curr_timestamp, ms_timestamp)

        self.enqueue_data(timestamp, data)

        self.last_ms = int(binascii.hexlify(data[-4:-2]), 16) % 2 ** 16  # recopy because ms_timestamp can be modified above
        self.last_timestamp = timestamp
        self.last_seq = seq

    def is_rpc(self, pid):
        return pid == SnapConstantsHelper.PID_RPC or pid == (SnapConstantsHelper.PID_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK)

    def is_multicast_rpc(self, pid):
        return pid == SnapConstantsHelper.PID_MULTICAST_RPC or pid == (SnapConstantsHelper.PID_MULTICAST_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK)

    def is_ack(self, pid):
        return pid == SnapConstantsHelper.PID_ACK or pid == SnapConstantsHelper.PID_NACK

    def is_mesh_ptp(self, pid):
        return pid == SnapConstantsHelper.PID_MESH_PTP or pid == (SnapConstantsHelper.PID_MESH_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK)

    def is_mesh_bcast(self, pid):
        return pid == SnapConstantsHelper.PID_MESH_BCAST or pid == (SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK)

    def is_dm_ptp(self, pid):
        return pid == SnapConstantsHelper.PID_DM_PTP or pid == (SnapConstantsHelper.PID_DM_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK)

    def is_dm_bcast(self, pid):
        return pid in (SnapConstantsHelper.PID_DM_BCAST, (SnapConstantsHelper.PID_DM_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK))

    def is_trace_route(self, pid):
        return pid in (SnapConstantsHelper.PID_TRACE_ROUTE, (SnapConstantsHelper.PID_TRACE_ROUTE | SnapConstantsHelper.PID_ENCRYPTION_MASK))

    def process_rpc(self, pid, data, seq):
        info = ""
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.FWD_UNI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.FWD_UNI_PKT_DATA:])

        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.RPC_ORIG_SRC_OFS:SnapConstantsHelper.RPC_ORIG_SRC_OFS + 3]))
        if (data[SnapConstantsHelper.RPC_ORIG_SRC_OFS:SnapConstantsHelper.RPC_ORIG_SRC_OFS + 3] !=
                data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]):
            src += " (%s)" % (binascii.hexlify(data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]))

        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.RPC_DSTADDR_OFS:SnapConstantsHelper.RPC_DSTADDR_OFS + 3]))
        if (data[SnapConstantsHelper.RPC_DSTADDR_OFS:SnapConstantsHelper.RPC_DSTADDR_OFS + 3] !=
                data[SnapConstantsHelper.RPC_HOPADDR_OFS:SnapConstantsHelper.RPC_HOPADDR_OFS + 3]):
            dst += " (%s)" % (binascii.hexlify(data[SnapConstantsHelper.RPC_HOPADDR_OFS:SnapConstantsHelper.RPC_HOPADDR_OFS + 3]))

        rpc = RpcCodec.RpcCodec.decode(data, checkLength=False)
        if rpc:
            info = "Method: %s%s" % (rpc.funcName.replace("\n", "\\n"), pprint.pformat(rpc.args).replace("\n", "\\n"))
            if rpc.recv_crc is not None:
                info += " CRC: %04X" % (rpc.recv_crc)
                if rpc.computed_crc is not None and rpc.recv_crc != rpc.computed_crc:
                    info += " (%04X)" % (rpc.computed_crc)
        return src, dst, info

    def process_multicast_rpc(self, pid, data, seq):
        info = ""
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_MULTICAST_RPC | 0x80) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.MULTI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.MULTI_PKT_DATA:])

        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MRPC_ORIG_SRC_OFS:SnapConstantsHelper.MRPC_ORIG_SRC_OFS + 3]))
        if (data[SnapConstantsHelper.MRPC_ORIG_SRC_OFS:SnapConstantsHelper.MRPC_ORIG_SRC_OFS + 3] !=
                data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]):
            src += " (%s)" % (binascii.hexlify(data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]))
        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.RPC_HOPADDR_OFS:SnapConstantsHelper.RPC_HOPADDR_OFS + 2]))
        dst += " TTL=%i " % (int(binascii.hexlify(data[SnapConstantsHelper.RPC_TTL_OFS:SnapConstantsHelper.RPC_TTL_OFS + 1]), 16))
        rpc = RpcCodec.RpcCodec.decode(data, checkLength=False)
        if rpc:
            info += "Method: %s%s" % (rpc.funcName.replace("\n", "\\n"), pprint.pformat(rpc.args).replace("\n", "\\n"))
            if rpc.recv_crc is not None:
                info += " CRC: %04X" % (rpc.recv_crc)
                if rpc.computed_crc is not None and rpc.recv_crc != rpc.computed_crc:
                    info += " (%04X)" % (rpc.computed_crc)
        return src, dst, info

    def process_ack(self, pid, data):
        info = ""
        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.ACK_SRC_OFS:SnapConstantsHelper.ACK_SRC_OFS + 3]))
        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.ACK_DST_OFS:SnapConstantsHelper.ACK_DST_OFS + 3]))
        if pid == SnapConstantsHelper.PID_NACK:
            pass
        return src, dst, info

    def process_mesh_ptp(self, pid, data, seq):
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_MESH_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.UNI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.UNI_PKT_DATA:])

        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MPTP_SRC_OFS:SnapConstantsHelper.MPTP_SRC_OFS + 3]))
        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MPTP_DST_OFS:SnapConstantsHelper.MPTP_DST_OFS + 3]))
        target = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_TGT_OFS:SnapConstantsHelper.MBST_TGT_OFS + 3]))
        orig = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_ORIG_OFS:SnapConstantsHelper.MBST_ORIG_OFS + 3]))
        type = data[SnapConstantsHelper.MPTP_TYP_OFS:SnapConstantsHelper.MPTP_TYP_OFS + 1]
        if type == 'P':
            info = "Type: RREP - %s responding to %s" % (orig, target)
        else:
            info = "Type: unknown"
        return src, dst, info

    def process_mesh_bcast(self, pid, data, seq):
        info = ""
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.MESH_MULTI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.MESH_MULTI_PKT_DATA:])

        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_GRP_OFS:SnapConstantsHelper.MBST_GRP_OFS + 2]))
        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_SRC_OFS:SnapConstantsHelper.MBST_SRC_OFS + 3]))
        target = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_TGT_OFS:SnapConstantsHelper.MBST_TGT_OFS + 3]))
        orig = "%s" % (binascii.hexlify(data[SnapConstantsHelper.MBST_ORIG_OFS:SnapConstantsHelper.MBST_ORIG_OFS + 3]))
        maxhops = int(binascii.hexlify(data[SnapConstantsHelper.MBST_MAXHOP_OFS]), 16)
        type = data[SnapConstantsHelper.MPTP_TYP_OFS:SnapConstantsHelper.MPTP_TYP_OFS + 1]
        if type == 'Q':
            info = "Type: RREQ - %s is looking for %s (MAXHOPS=%d)" % (orig, target, maxhops)
        elif type == 'E':
            info = "Type: RERR - %s (MAXHOPS=%d)" % (target, maxhops)
        return src, dst, info

    def process_dm_ptp(self, pid, data, seq):
        info = ""
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_DM_PTP | 0x80) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.FWD_UNI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.FWD_UNI_PKT_DATA:])
        src = "%s (%s)" % (binascii.hexlify(data[SnapConstantsHelper.DMPTP_ORIG_SRC_OFS:SnapConstantsHelper.DMPTP_ORIG_SRC_OFS + 3]),
                           binascii.hexlify(data[SnapConstantsHelper.DMPTP_SRC_OFS:SnapConstantsHelper.DMPTP_SRC_OFS + 3]))
        dst = "%s (%s)" % (binascii.hexlify(data[SnapConstantsHelper.DMPTP_FINALDST_OFS:SnapConstantsHelper.DMPTP_FINALDST_OFS + 3]),
                           binascii.hexlify(data[SnapConstantsHelper.DMPTP_DST_OFS:SnapConstantsHelper.DMPTP_DST_OFS + 3]))
        info += "%s" % (pprint.pformat(data[SnapConstantsHelper.DMPTP_ORIG_SRC_OFS + 3:-2]))
        return src, dst, info

    def process_dm_bcast(self, pid, data, seq):
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_DM_BCAST | 0x80) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.MULTI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.MULTI_PKT_DATA:])

        src = "%s" % (binascii.hexlify(data[SnapConstantsHelper.DMBCAST_ORIG_SRC_OFS:SnapConstantsHelper.DMBCAST_ORIG_SRC_OFS + SnapConstantsHelper.SNAP2_ADDR_LEN]))
        if (binascii.hexlify(data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]) !=
                binascii.hexlify(data[SnapConstantsHelper.DMBCAST_ORIG_SRC_OFS:SnapConstantsHelper.DMBCAST_ORIG_SRC_OFS + SnapConstantsHelper.SNAP2_ADDR_LEN])):
            src += " (%s)" % (binascii.hexlify(data[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS + 3]))
        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.DMPTP_DST_OFS:SnapConstantsHelper.DMPTP_DST_OFS + 2]))
        info = "TTL=%i " % (int(binascii.hexlify(data[SnapConstantsHelper.DMBAST_TTL_OFS:SnapConstantsHelper.DMBAST_TTL_OFS + 1]), 16))
        info += "Data: %s" % (pprint.pformat(data[SnapConstantsHelper.DMPTP_ORIG_SRC_OFS:]))
        return src, dst, info

    def process_trace_route(self, pid, data, seq):
        if AES_SUPPORT and pid == (SnapConstantsHelper.PID_TRACE_ROUTE | 0x80) and self.crypto.has_valid_key:
            data = data[:SnapConstantsHelper.FWD_UNI_PKT_DATA] + self.crypto.decrypt(int(seq, 16), data[SnapConstantsHelper.FWD_UNI_PKT_DATA:])

        src = "%s (%s)" % (binascii.hexlify(data[SnapConstantsHelper.TR_ORIG_SRC_OFS:SnapConstantsHelper.TR_ORIG_SRC_OFS + 3]),
                           binascii.hexlify(data[SnapConstantsHelper.TR_SRC_OFS:SnapConstantsHelper.TR_SRC_OFS + 3]))
        dst = "%s" % (binascii.hexlify(data[SnapConstantsHelper.TR_DST_OFS:SnapConstantsHelper.TR_DST_OFS + 3]))
        desc = TraceRouteCodec.decode(data)
        # info = ""
        # How much of the following should we show the end users? I used it for initial Trace Route testing, but maybe we want to comment this out?
        info = desc.msgId + ' ' + str(desc.sequence) + ' ' + str(desc.elapsed) + ' '
        for record in desc.records:
            assert isinstance(record, TraceRouteCodec.TraceRouteRecord)
            info += "%s=%i;" % (binascii.hexlify(record.address), record.linkQuality)
        return src, dst, info

    def process_data(self, timestamp, data, lq):
        seq = 'Unknown'
        pid = 'Unknown'
        pktLen = 'Unknown'
        src = None
        dst = None
        info = None

        try:
            pid = int(binascii.hexlify(data[SnapConstantsHelper.SNAP_SNYF_PID_OFS]), 16)
            pktLen = int(binascii.hexlify(data[SnapConstantsHelper.SNAP_SNYF_LEN_OFS]), 16)
            seq = binascii.hexlify(data[SnapConstantsHelper.RPC_SEQ_OFS])

            if self.is_rpc(pid):
                src, dst, info = self.process_rpc(pid, data, seq)
            elif self.is_multicast_rpc(pid):
                src, dst, info = self.process_multicast_rpc(pid, data, seq)
            elif self.is_ack(pid):
                src, dst, info = self.process_ack(pid, data)
            elif self.is_mesh_ptp(pid):
                src, dst, info = self.process_mesh_ptp(pid, data, seq)
            elif self.is_mesh_bcast(pid):
                src, dst, info = self.process_mesh_bcast(pid, data, seq)
            elif self.is_dm_ptp(pid):
                src, dst, info = self.process_dm_ptp(pid, data, seq)
            elif self.is_dm_bcast(pid):
                src, dst, info = self.process_dm_bcast(pid, data, seq)
            elif self.is_trace_route(pid):
                src, dst, info = self.process_trace_route(pid, data, seq)
        except:
            log.exception("Error during decode:")

        event = SnifferEvent(timestamp=timestamp,
                             link_quality=lq,
                             packet_length=pktLen,
                             seq=seq,
                             source=src,
                             destination=dst,
                             packet_type=SnapConstantsHelper.getPidString(pid),
                             info=info,
                             data=data)

        for cb in self.event_callbacks:
            try:
                tornado.ioloop.IOLoop.current().add_callback(cb, event)
            except:
                log.exception("Sniffer event callback")

    def open(self, port):
        """Open a connection."""
        log.debug("Attempting to set the output type")
        if self.pyserial_driver.isOpen:
            self.pyserial_driver.close()
        try:
            self.pyserial_driver.setOutputType(PyserialDriver.PyserialWrapper.TYPE_PYSERIAL, port)
        except PyserialDriver.SerialOpenException:
            raise SnifferException('Unable to open serial port: %s' % port)

    def start(self, channel):
        log.debug("Start sniffing")
        if channel < 0 or channel > 15:
            raise ValueError
        self.pkt_serial.send_data("%s%s" % (START_CMD, binascii.unhexlify("%02X" % channel)))

    def stop(self):
        self.pkt_serial.send_data(STOP_CMD)

    def set_encryption(self, snap_encryption_type, snap_encryption_key):
        if snap_encryption_type == SnapConstantsHelper.ENCRYPTION_TYPE_AES128:
            self.crypto_enabled = True
            self.crypto_type = SnapConstantsHelper.ENCRYPTION_TYPE_AES128
            self.crypto_key = snap_encryption_key
            self.crypto = aes.CipherStealingAES128(self.crypto_key)
        elif snap_encryption_type == SnapConstantsHelper.ENCRYPTION_TYPE_BASIC:
            self.crypto_enabled = True
            self.crypto_type = SnapConstantsHelper.ENCRYPTION_TYPE_BASIC
            self.crypto_key = snap_encryption_key
            self.crypto = pme.PoorManEncryption(self.crypto_key)
        elif snap_encryption_type == SnapConstantsHelper.ENCRYPTION_TYPE_NONE:
            self.crypto_enabled = False
            self.crypto_type = SnapConstantsHelper.ENCRYPTION_TYPE_NONE
            self.crypto_key = snap_encryption_key
            self.crypto = aes.CipherStealingAES128("")
        else:
            raise SnifferException("Unrecognized SNAP encryption type: %s" % snap_encryption_type)

        Snap2Decoder.crypto = self.crypto
