"""SNAPstraction Utility Functions"""
# Future SNAP Connect Context Manager
import logging

import binascii
import re

log = logging.getLogger(__name__)


ADDR_RE = re.compile(r'^[0-9A-Fa-f]{6}$')


CRYPTO_LOOKUP = {0: "No",
                 1: "AES-128",
                 2: "Basic"}


def first_or_none(result):
    if isinstance(result, list):
        result = tuple(result)

    if result is not None and isinstance(result, tuple):
        log.debug("Got a tuple result: %r", result)
        return result[0]
    else:
        log.debug("Got a non-tuple result: %r", result)
        return result


def parse_value_for_snap(string):
    # Do RCMD-style parsing of the value
    try:
        return int(string, 0)
    except:
        if string and len(string) > 0 and string[0] == 'x':
            return binascii.unhexlify(string[1:])

        # Is it an empty string?
        if string == "''" or string == '""' or string == "":
            return ""

        if string[0] == "'" and string[-1] == "'" and len(string) > 2:
            return string[1:-1]

        if string[0] == '"' and string[-1] == '"' and len(string) > 2:
            return string[1:-1]

        if string.lower() == "true":
            return True
        if string.lower() == "false":
            return False
        if string.lower() == "none":
            return None
    return string


def validate_snap_address(addr):
    if not isinstance(addr, basestring):
        return None

    address = addr.lower()

    if address == 'bridge' or (ADDR_RE.match(address) is not None):
        return address
    else:
        return None


def clean_up(t, bridge):
    """If the target is 'bridge', replace it with the bridge address."""
    t = validate_snap_address(t)
    if t == 'bridge':
        t = bridge

    return t


def fix_up_target(func):
    """Wrap the func with "clean up targets first"

    Args:
        func (function): function to wrap with "clean up/replace target"

    Returns:
        Wrapped function.

    """

    def fixed(gus, data, target, *args, **kwargs):
        # Clean up iterable of targets, too.
        if isinstance(target, basestring):
            target = clean_up(target, gus.bridge_address)
        else:
            try:
                target = iter(target)
            except:
                raise Exception("%r was not a valid target" % target)
            else:
                target = [clean_up(t, gus.bridge_address) for t in target]

        return func(gus, data, target, *args, **kwargs)

    return fixed
